@@ -3,6 +3,7 @@
 #include <config.h>
 #endif
 
+#include <gio/gio.h>
 #include <ibus.h>
 #include <m17n.h>
 #include <string.h>
@@ -23,28 +24,30 @@ struct _IBusM17NEngine {
     IBusProperty    *setup_prop;
 #endif  /* HAVE_SETUP */
     IBusPropList    *prop_list;
+    IBusKeymap      *us_keymap;
+    IBusInputPurpose purpose;
+    IBusInputHints   hints;
 };
 
 struct _IBusM17NEngineClass {
     IBusEngineClass parent;
 
     /* configurations are per class */
-    gchar *config_section;
+    GSettings *gsettings;
     guint preedit_foreground;
     guint preedit_background;
     gint preedit_underline;
     IBusPreeditFocusMode preedit_focus_mode;
     gint lookup_table_orientation;
+    gboolean use_us_layout;
 
     MInputMethod *im;
 };
 
 /* functions prototype */
 static void ibus_m17n_engine_class_init     (IBusM17NEngineClass    *klass);
-static void ibus_m17n_config_value_changed  (IBusConfig             *config,
-                                             const gchar            *section,
-                                             const gchar            *name,
-                                             GVariant               *value,
+static void ibus_m17n_config_value_changed  (GSettings              *gsettings,
+                                             const gchar            *key,
                                              IBusM17NEngineClass    *klass);
 
 static GObject*
@@ -64,14 +67,6 @@ static void ibus_m17n_engine_focus_out      (IBusEngine             *engine);
 static void ibus_m17n_engine_reset          (IBusEngine             *engine);
 static void ibus_m17n_engine_enable         (IBusEngine             *engine);
 static void ibus_m17n_engine_disable        (IBusEngine             *engine);
-static void ibus_engine_set_cursor_location (IBusEngine             *engine,
-                                             gint                    x,
-                                             gint                    y,
-                                             gint                    w,
-                                             gint                    h);
-static void ibus_m17n_engine_set_capabilities
-                                            (IBusEngine             *engine,
-                                             guint                   caps);
 static void ibus_m17n_engine_page_up        (IBusEngine             *engine);
 static void ibus_m17n_engine_page_down      (IBusEngine             *engine);
 static void ibus_m17n_engine_cursor_up      (IBusEngine             *engine);
@@ -80,12 +75,10 @@ static void ibus_m17n_engine_property_activate
                                             (IBusEngine             *engine,
                                              const gchar            *prop_name,
                                              guint                   prop_state);
-static void ibus_m17n_engine_property_show
+static void ibus_m17n_engine_set_content_type
                                             (IBusEngine             *engine,
-                                             const gchar            *prop_name);
-static void ibus_m17n_engine_property_hide
-                                            (IBusEngine             *engine,
-                                             const gchar            *prop_name);
+                                             IBusInputPurpose        purpose,
+                                             IBusInputHints          hints);
 
 static void ibus_m17n_engine_commit_string
                                             (IBusM17NEngine         *m17n,
@@ -98,14 +91,9 @@ static void ibus_m17n_engine_update_lookup_table
 
 static IBusEngineClass *parent_class = NULL;
 
-static IBusConfig      *config = NULL;
-
 void
 ibus_m17n_init (IBusBus *bus)
 {
-    config = ibus_bus_get_config (bus);
-    if (config)
-        g_object_ref_sink (config);
     ibus_m17n_init_common ();
 }
 
@@ -212,7 +200,6 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
     IBusEngineClass *engine_class = IBUS_ENGINE_CLASS (klass);
     gchar *engine_name, *lang = NULL, *name = NULL;
     IBusM17NEngineConfig *engine_config;
-    GVariant *values;
 
     if (parent_class == NULL)
         parent_class = (IBusEngineClass *) g_type_class_peek_parent (klass);
@@ -237,17 +224,22 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
 
     engine_class->property_activate = ibus_m17n_engine_property_activate;
 
+    engine_class->set_content_type = ibus_m17n_engine_set_content_type;
+
     if (!ibus_m17n_scan_class_name (G_OBJECT_CLASS_NAME (klass),
                                     &lang, &name)) {
         g_free (lang);
         g_free (name);
         return;
     }
-    engine_name = g_strdup_printf ("m17n:%s:%s", lang, name);
-    klass->config_section = g_strdup_printf ("engine/M17N/%s/%s", lang, name);
+    klass->gsettings = g_settings_new_with_path (
+        "org.freedesktop.ibus.engine.m17n",
+        g_strdup_printf ("/org/freedesktop/ibus/engine/m17n/%s/%s/",
+                         lang, name));
     g_free (lang);
     g_free (name);
 
+    engine_name = g_strdup_printf ("m17n:%s:%s", lang, name);
     engine_config = ibus_m17n_get_engine_config (engine_name);
     g_free (engine_name);
 
@@ -263,51 +255,52 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
         IBUS_ENGINE_PREEDIT_COMMIT :
         IBUS_ENGINE_PREEDIT_CLEAR;
     klass->lookup_table_orientation = IBUS_ORIENTATION_SYSTEM;
+    klass->use_us_layout = FALSE;
 
     ibus_m17n_engine_config_free (engine_config);
 
-    values = ibus_config_get_values (config,
-                                     klass->config_section);
-    if (values != NULL) {
+    if (klass->gsettings != NULL) {
         GVariant *value;
 
-        value = g_variant_lookup_value (values,
-                                        "preedit_foreground",
-                                        G_VARIANT_TYPE_STRING);
+        value = g_settings_get_value (klass->gsettings,
+                                      "preedit-foreground");
         if (value != NULL) {
             const gchar *hex = g_variant_get_string (value, NULL);
             klass->preedit_foreground = ibus_m17n_parse_color (hex);
             g_variant_unref (value);
         }
 
-        value = g_variant_lookup_value (values,
-                                        "preedit_background",
-                                        G_VARIANT_TYPE_STRING);
+        value = g_settings_get_value (klass->gsettings,
+                                      "preedit-background");
         if (value != NULL) {
             const gchar *hex = g_variant_get_string (value, NULL);
             klass->preedit_background = ibus_m17n_parse_color (hex);
             g_variant_unref (value);
         }
 
-        value = g_variant_lookup_value (values,
-                                        "preedit_underline",
-                                        G_VARIANT_TYPE_INT32);
+        value = g_settings_get_value (klass->gsettings,
+                                      "preedit-underline");
         if (value != NULL) {
-            klass->preedit_background = g_variant_get_int32 (value);
+            klass->preedit_underline = g_variant_get_int32 (value);
             g_variant_unref (value);
         }
 
-        value = g_variant_lookup_value (values,
-                                        "lookup_table_orientation",
-                                        G_VARIANT_TYPE_INT32);
+        value = g_settings_get_value (klass->gsettings,
+                                      "lookup-table-orientation");
         if (value != NULL) {
             klass->lookup_table_orientation = g_variant_get_int32 (value);
             g_variant_unref (value);
         }
-        g_variant_unref (values);
+
+        value = g_settings_get_value (klass->gsettings,
+                                      "use-us-layout");
+        if (value != NULL) {
+            klass->use_us_layout = g_variant_get_boolean (value);
+            g_variant_unref (value);
+        }
     }
 
-    g_signal_connect (config, "value-changed",
+    g_signal_connect (klass->gsettings, "changed",
                       G_CALLBACK(ibus_m17n_config_value_changed),
                       klass);
 
@@ -315,33 +308,40 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
 }
 
 static void
-ibus_m17n_config_value_changed (IBusConfig          *config,
-                                const gchar         *section,
-                                const gchar         *name,
-                                GVariant            *value,
+ibus_m17n_config_value_changed (GSettings           *gsettings,
+                                const gchar         *key,
                                 IBusM17NEngineClass *klass)
 {
-    if (g_strcmp0 (section, klass->config_section) == 0) {
-        if (g_strcmp0 (name, "preedit_foreground") == 0) {
-            const gchar *hex = g_variant_get_string (value, NULL);
-            guint color;
-            color = ibus_m17n_parse_color (hex);
-            if (color != INVALID_COLOR) {
-                klass->preedit_foreground = color;
-            }
-        } else if (g_strcmp0 (name, "preedit_background") == 0) {
-            const gchar *hex = g_variant_get_string (value, NULL);
-            guint color;
-            color = ibus_m17n_parse_color (hex);
-            if (color != INVALID_COLOR) {
-                klass->preedit_background = color;
-            }
-        } else if (g_strcmp0 (name, "preedit_underline") == 0) {
-            klass->preedit_underline = g_variant_get_int32 (value);
-        } else if (g_strcmp0 (name, "lookup_table_orientation") == 0) {
-            klass->lookup_table_orientation = g_variant_get_int32 (value);
+    GVariant *value;
+    value = g_settings_get_value (gsettings, key);
+    if (g_strcmp0 (key, "preedit-foreground") == 0) {
+        const gchar *hex = g_variant_get_string (value, NULL);
+        guint color;
+        color = ibus_m17n_parse_color (hex);
+        if (color != INVALID_COLOR) {
+            klass->preedit_foreground = color;
         }
+        else {
+            klass->preedit_foreground = INVALID_COLOR;
+        }
+    } else if (g_strcmp0 (key, "preedit-background") == 0) {
+        const gchar *hex = g_variant_get_string (value, NULL);
+        guint color;
+        color = ibus_m17n_parse_color (hex);
+        if (color != INVALID_COLOR) {
+            klass->preedit_background = color;
+        }
+        else {
+            klass->preedit_background = INVALID_COLOR;
+        }
+    } else if (g_strcmp0 (key, "preedit-underline") == 0) {
+        klass->preedit_underline = g_variant_get_int32 (value);
+    } else if (g_strcmp0 (key, "lookup-table-orientation") == 0) {
+        klass->lookup_table_orientation = g_variant_get_int32 (value);
+    } else if (g_strcmp0 (key, "use-us-layout") == 0) {
+        klass->use_us_layout = g_variant_get_boolean (value);
     }
+    g_variant_unref (value);
 }
 
 static void
@@ -384,6 +384,7 @@ ibus_m17n_engine_init (IBusM17NEngine *m17n)
     m17n->table = ibus_lookup_table_new (9, 0, TRUE, TRUE);
     g_object_ref_sink (m17n->table);
     m17n->context = NULL;
+    m17n->us_keymap = ibus_keymap_get ("us");
 }
 
 static GObject*
@@ -476,6 +477,11 @@ ibus_m17n_engine_destroy (IBusM17NEngine *m17n)
         m17n->context = NULL;
     }
 
+    if (m17n->us_keymap) {
+        g_object_unref (m17n->us_keymap);
+        m17n->us_keymap = NULL;
+    }
+
     IBUS_OBJECT_CLASS (parent_class)->destroy ((IBusObject *)m17n);
 }
 
@@ -531,28 +537,25 @@ ibus_m17n_engine_commit_string (IBusM17NEngine *m17n,
    Since IBus engines are supposed to be cross-platform, the code
    should go into IBus core, instead of ibus-m17n. */
 static MSymbol
-ibus_m17n_key_event_to_symbol (guint keycode,
-                               guint keyval,
-                               guint modifiers)
+ibus_m17n_key_event_to_symbol (IBusM17NEngine *m17n,
+                               guint           keycode,
+                               guint           keyval,
+                               guint           modifiers)
 {
     GString *keysym;
     MSymbol mkeysym = Mnil;
     guint mask = 0;
-    IBusKeymap *keymap;
 
     if (keyval >= IBUS_Shift_L && keyval <= IBUS_Hyper_R) {
         return Mnil;
     }
 
-    /* Here, keyval is already translated by IBUS_MOD5_MASK.  Obtain
-       the untranslated keyval from the underlying keymap and
-       represent the translated keyval as the form "G-<untranslated
-       keyval>", which m17n-lib accepts. */
+    /* If keyval is already translated by IBUS_MOD5_MASK.  Try to
+       obtain the untranslated keyval from the US keymap. */
     if (modifiers & IBUS_MOD5_MASK) {
-        keymap = ibus_keymap_get ("us");
-        keyval = ibus_keymap_lookup_keysym (keymap, keycode,
+        keyval = ibus_keymap_lookup_keysym (m17n->us_keymap,
+                                            keycode,
                                             modifiers & ~IBUS_MOD5_MASK);
-        g_object_unref (keymap);
     }
 
     keysym = g_string_new ("");
@@ -656,15 +659,49 @@ ibus_m17n_engine_process_key_event (IBusEngine     *engine,
                                     guint           modifiers)
 {
     IBusM17NEngine *m17n = (IBusM17NEngine *) engine;
+    IBusM17NEngineClass *klass =
+        (IBusM17NEngineClass *) G_OBJECT_GET_CLASS (m17n);
+    guint original_keyval = keyval;
+
+    switch (m17n->purpose) {
+    case IBUS_INPUT_PURPOSE_PASSWORD:
+    case IBUS_INPUT_PURPOSE_PIN:
+        /* For password and PIN input, skip any further step
+           processing a key event.  */
+        return FALSE;
+
+    default:
+        break;
+    }
 
     if (modifiers & IBUS_RELEASE_MASK)
         return FALSE;
-    MSymbol m17n_key = ibus_m17n_key_event_to_symbol (keycode, keyval, modifiers);
 
-    if (m17n_key == Mnil)
-        return FALSE;
+    if (klass->use_us_layout) {
+        keyval = ibus_keymap_lookup_keysym (m17n->us_keymap,
+                                            keycode,
+                                            modifiers);
+    }
 
-    return ibus_m17n_engine_process_key (m17n, m17n_key);
+    MSymbol m17n_key = ibus_m17n_key_event_to_symbol (m17n,
+                                                      keycode,
+                                                      keyval,
+                                                      modifiers);
+    if (m17n_key != Mnil && ibus_m17n_engine_process_key (m17n, m17n_key)) {
+        return TRUE;
+    }
+
+    /* If keyval is translated in US layout, send the new keyval and
+       notify that the event is handled. */
+    if (keyval != original_keyval && 0x20 <= keyval && keyval < 0x7F) {
+        gchar buf[2];
+        buf[0] = keyval;
+        buf[1] = '\0';
+        ibus_m17n_engine_commit_string (m17n, buf);
+        return TRUE;
+    }
+
+    return FALSE;
 }
 
 static void
@@ -704,8 +741,6 @@ ibus_m17n_engine_reset (IBusEngine *engine)
 static void
 ibus_m17n_engine_enable (IBusEngine *engine)
 {
-    IBusM17NEngine *m17n = (IBusM17NEngine *) engine;
-
     parent_class->enable (engine);
 
     /* Issue a dummy ibus_engine_get_surrounding_text() call to tell
@@ -716,8 +751,6 @@ ibus_m17n_engine_enable (IBusEngine *engine)
 static void
 ibus_m17n_engine_disable (IBusEngine *engine)
 {
-    IBusM17NEngine *m17n = (IBusM17NEngine *) engine;
-
     ibus_m17n_engine_focus_out (engine);
     parent_class->disable (engine);
 }
@@ -785,6 +818,31 @@ ibus_m17n_engine_property_activate (IBusEngine  *engine,
     parent_class->property_activate (engine, prop_name, prop_state);
 }
 
+static void
+ibus_m17n_engine_set_content_type (IBusEngine      *engine,
+                                   IBusInputPurpose purpose,
+                                   IBusInputHints   hints)
+{
+    IBusM17NEngine *m17n = (IBusM17NEngine *) engine;
+
+    m17n->purpose = purpose;
+    m17n->hints = hints;
+
+    switch (purpose) {
+    case IBUS_INPUT_PURPOSE_PASSWORD:
+    case IBUS_INPUT_PURPOSE_PIN:
+        /* For password and PIN input, emulate 'focus-out' to discard
+           any pending input status (e.g. preedit or candidate
+           list).  */
+        ibus_m17n_engine_focus_out (engine);
+        break;
+
+    default:
+        ibus_m17n_engine_focus_in (engine);
+        break;
+    }
+}
+
 static void
 ibus_m17n_engine_update_lookup_table (IBusM17NEngine *m17n)
 {
@@ -951,7 +1009,9 @@ ibus_m17n_engine_callback (MInputContext *context,
                                           &anchor_pos);
         nchars = ibus_text_get_length (text);
         nbytes = g_utf8_offset_to_pointer (text->text, nchars) - text->text;
-        mt = mconv_decode_buffer (Mcoding_utf_8, text->text, nbytes);
+        mt = mconv_decode_buffer (Mcoding_utf_8,
+                                  (const unsigned char *) text->text,
+                                  nbytes);
         g_object_unref (text);
 
         len = (long) mplist_value (m17n->context->plist);
