@@ -11,14 +11,21 @@
 
 static MConverter *utf8_converter = NULL;
 
-#define DEFAULT_XML (SETUPDIR "/default.xml")
+#define DEFAULT_XML (PKGDATADIR "/default.xml")
 
-struct _IBusM17NEngineConfigNode {
+typedef enum {
+    ENGINE_CONFIG_RANK_MASK = 1 << 0,
+    ENGINE_CONFIG_SYMBOL_MASK = 1 << 1,
+    ENGINE_CONFIG_LONGNAME_MASK = 1 << 2,
+    ENGINE_CONFIG_PREEDIT_HIGHLIGHT_MASK = 1 << 3
+} EngineConfigMask;
+
+struct _EngineConfigNode {
     gchar *name;
+    EngineConfigMask mask;
     IBusM17NEngineConfig config;
 };
-
-typedef struct _IBusM17NEngineConfigNode IBusM17NEngineConfigNode;
+typedef struct _EngineConfigNode EngineConfigNode;
 
 static GSList *config_list = NULL;
 
@@ -46,7 +53,9 @@ ibus_m17n_mtext_to_utf8 (MText *text)
     bufsize = (mtext_len (text) + 1) * 6;
     buf = (gchar *) g_malloc (bufsize);
 
-    mconv_rebind_buffer (utf8_converter, buf, bufsize);
+    mconv_rebind_buffer (utf8_converter,
+                         (const unsigned char *) buf,
+                         bufsize);
     mconv_encode (utf8_converter, text);
 
     buf [utf8_converter->nbytes] = 0;
@@ -69,7 +78,9 @@ ibus_m17n_mtext_to_ucs4 (MText *text, glong *nchars)
     bufsize = (mtext_len (text) + 1) * 6;
     buf = (gchar *) g_malloc (bufsize);
 
-    mconv_rebind_buffer (utf8_converter, buf, bufsize);
+    mconv_rebind_buffer (utf8_converter,
+                         (const unsigned char *) buf,
+                         bufsize);
     if (mconv_encode (utf8_converter, text) < 0) {
         g_free (buf);
         return NULL;
@@ -109,6 +120,7 @@ ibus_m17n_engine_new (MSymbol  lang,
     gchar *engine_title;
     gchar *engine_icon;
     gchar *engine_desc;
+    gchar *engine_setup;
 
     engine_name = g_strdup_printf ("m17n:%s:%s", msymbol_name (lang), msymbol_name (name));
 
@@ -116,34 +128,27 @@ ibus_m17n_engine_new (MSymbol  lang,
     engine_title = ibus_m17n_mtext_to_utf8 (title);
     engine_icon = ibus_m17n_mtext_to_utf8 (icon);
     engine_desc = ibus_m17n_mtext_to_utf8 (desc);
+    engine_setup = g_strdup_printf ("%s/ibus-setup-m17n --name %s",
+                                    LIBEXECDIR, engine_name);
 
-#if IBUS_CHECK_VERSION(1,3,99)
     engine = ibus_engine_desc_new_varargs ("name",        engine_name,
-                                           "longname",    engine_longname,
+                                           "longname",    config->longname ? config->longname : engine_longname,
                                            "description", engine_desc ? engine_desc : "",
                                            "language",    msymbol_name (lang),
                                            "license",     "GPL",
                                            "icon",        engine_icon ? engine_icon : "",
-                                           "layout",      "us",
+                                           "layout",      "default",
                                            "rank",        config->rank,
+                                           "symbol",      config->symbol ? config->symbol : "",
+                                           "setup",       engine_setup,
                                            NULL);
-#else
-    engine = ibus_engine_desc_new (engine_name,
-                                   engine_longname,
-                                   engine_desc ? engine_desc : "",
-                                   msymbol_name (lang),
-                                   "GPL",
-                                   "",
-                                   engine_icon ? engine_icon : "",
-                                   "us");
-    engine->rank = config->rank;
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
 
     g_free (engine_name);
     g_free (engine_longname);
     g_free (engine_title);
     g_free (engine_icon);
     g_free (engine_desc);
+    g_free (engine_setup);
 
     return engine;
 }
@@ -188,8 +193,10 @@ ibus_m17n_list_engines (void)
                 continue;
             }
             if (config->rank < 0) {
-                g_warning ("skipped %s since its rank is lower than 0",
-                           engine_name);
+                g_log ("ibus-m17n",
+                       G_LOG_LEVEL_MESSAGE,
+                       "skipped %s since its rank is lower than 0",
+                       engine_name);
                 g_free (engine_name);
                 continue;
             }
@@ -199,17 +206,12 @@ ibus_m17n_list_engines (void)
             if (l) {
                 /* check candidates encoding */
                 MPlist *sl;
-                MSymbol varname;
-                MText *vardesc;
-                MSymbol varunknown;
                 MSymbol varcharset;
 
                 sl = mplist_value (l);
-                varname  = mplist_value (sl);
+                /* L = (VAR-NAME DESCRIPTION 'nil' VALUE) */
                 sl = mplist_next (sl);
-                vardesc = mplist_value (sl);
                 sl = mplist_next (sl);
-                varunknown = mplist_value (sl);
                 sl = mplist_next (sl);
                 varcharset = mplist_value (sl);
 
@@ -257,20 +259,35 @@ ibus_m17n_list_engines (void)
 IBusM17NEngineConfig *
 ibus_m17n_get_engine_config (const gchar *engine_name)
 {
+    IBusM17NEngineConfig *config = g_slice_new0 (IBusM17NEngineConfig);
     GSList *p;
 
     for (p = config_list; p != NULL; p = p->next) {
-        IBusM17NEngineConfigNode *cnode = p->data;
+        EngineConfigNode *cnode = p->data;
 
-        if (g_pattern_match_simple (cnode->name, engine_name))
-            return &cnode->config;
+        if (g_pattern_match_simple (cnode->name, engine_name)) {
+            if (cnode->mask & ENGINE_CONFIG_RANK_MASK)
+                config->rank = cnode->config.rank;
+            if (cnode->mask & ENGINE_CONFIG_SYMBOL_MASK)
+                config->symbol = cnode->config.symbol;
+            if (cnode->mask & ENGINE_CONFIG_LONGNAME_MASK)
+                config->longname = cnode->config.longname;
+            if (cnode->mask & ENGINE_CONFIG_PREEDIT_HIGHLIGHT_MASK)
+                config->preedit_highlight = cnode->config.preedit_highlight;
+        }
     }
-    g_return_val_if_reached (NULL);
+    return config;
+}
+
+void
+ibus_m17n_engine_config_free (IBusM17NEngineConfig *config)
+{
+    g_slice_free (IBusM17NEngineConfig, config);
 }
 
 static gboolean
-ibus_m17n_engine_config_parse_xml_node (IBusM17NEngineConfigNode *cnode,
-                                        XMLNode                  *node)
+ibus_m17n_engine_config_parse_xml_node (EngineConfigNode *cnode,
+                                        XMLNode          *node)
 {
     GList *p;
 
@@ -284,6 +301,17 @@ ibus_m17n_engine_config_parse_xml_node (IBusM17NEngineConfigNode *cnode,
         }
         if (g_strcmp0 (sub_node->name , "rank") == 0) {
             cnode->config.rank = atoi (sub_node->text);
+            cnode->mask |= ENGINE_CONFIG_RANK_MASK;
+            continue;
+        }
+        if (g_strcmp0 (sub_node->name , "symbol") == 0) {
+            cnode->config.symbol = g_strdup (sub_node->text);
+            cnode->mask |= ENGINE_CONFIG_SYMBOL_MASK;
+            continue;
+        }
+        if (g_strcmp0 (sub_node->name , "longname") == 0) {
+            cnode->config.longname = g_strdup (sub_node->text);
+            cnode->mask |= ENGINE_CONFIG_LONGNAME_MASK;
             continue;
         }
         if (g_strcmp0 (sub_node->name , "preedit-highlight") == 0) {
@@ -292,6 +320,7 @@ ibus_m17n_engine_config_parse_xml_node (IBusM17NEngineConfigNode *cnode,
             else if (g_ascii_strcasecmp ("FALSE", sub_node->text) != 0)
                 g_warning ("<%s> element contains invalid boolean value %s",
                            sub_node->name, sub_node->text);
+            cnode->mask |= ENGINE_CONFIG_PREEDIT_HIGHLIGHT_MASK;
             continue;
         }
         g_warning ("<engine> element contains invalid element <%s>",
@@ -306,21 +335,28 @@ ibus_m17n_get_component (void)
     GList *engines, *p;
     IBusComponent *component;
     XMLNode *node;
+    const gchar *pkgdatadir;
+    gchar *default_xml;
 
     component = ibus_component_new ("org.freedesktop.IBus.M17n",
                                     N_("M17N"),
-                                    "0.1.0",
+                                    PACKAGE_VERSION,
                                     "GPL",
                                     "Peng Huang <shawn.p.huang@gmail.com>",
                                     "http://code.google.com/p/ibus/",
                                     "",
-                                    "ibus-m17n");
+                                    PACKAGE_NAME);
 
-    node = ibus_xml_parse_file (DEFAULT_XML);
+    pkgdatadir = g_getenv ("IBUS_M17N_PKGDATADIR");
+    if (pkgdatadir == NULL)
+        pkgdatadir = PKGDATADIR;
+    default_xml = g_build_filename (pkgdatadir, "default.xml", NULL);
+
+    node = ibus_xml_parse_file (default_xml);
     if (node && g_strcmp0 (node->name, "engines") == 0) {
         for (p = node->sub_nodes; p != NULL; p = p->next) {
             XMLNode *sub_node = p->data;
-            IBusM17NEngineConfigNode *cnode;
+            EngineConfigNode *cnode;
 
             if (g_strcmp0 (sub_node->name, "engine") != 0) {
                 g_warning ("<engines> element contains invalid element <%s>",
@@ -328,19 +364,21 @@ ibus_m17n_get_component (void)
                 continue;
             }
 
-            cnode = g_slice_new0 (IBusM17NEngineConfigNode);
+            cnode = g_slice_new0 (EngineConfigNode);
             if (!ibus_m17n_engine_config_parse_xml_node (cnode, sub_node)) {
-                g_slice_free (IBusM17NEngineConfigNode, cnode);
+                g_slice_free (EngineConfigNode, cnode);
                 continue;
             }
             config_list = g_slist_prepend (config_list, cnode);
         }
         config_list = g_slist_reverse (config_list);
     } else
-        g_warning ("failed to parse %s", DEFAULT_XML);
+        g_warning ("failed to parse %s", default_xml);
     if (node)
         ibus_xml_free (node);
 
+    g_free (default_xml);
+
     engines = ibus_m17n_list_engines ();
 
     for (p = engines; p != NULL; p = p->next)
@@ -350,128 +388,3 @@ ibus_m17n_get_component (void)
 
     return component;
 }
-
-void
-ibus_m17n_config_set_string (IBusConfig  *config,
-                             const gchar *section,
-                             const gchar *name,
-                             const gchar *value)
-{
-#if IBUS_CHECK_VERSION(1,3,99)
-    ibus_config_set_value (config, section, name, g_variant_new_string (value));
-#else
-    GValue v = { 0 };
-
-    g_value_init (&v, G_TYPE_STRING);
-    g_value_set_string (&v, value);
-    ibus_config_set_value (config, section, name, &v);
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
-}
-
-gboolean
-ibus_m17n_config_get_string (IBusConfig  *config,
-                             const gchar *section,
-                             const gchar *name,
-                             gchar      **result)
-{
-#if IBUS_CHECK_VERSION(1,3,99)
-    GVariant *value = NULL;
-
-    g_return_val_if_fail (result != NULL, FALSE);
-
-    value = ibus_config_get_value (config, section, name);
-    if (value) {
-        *result = g_strdup (g_variant_get_string (value, NULL));
-        g_variant_unref (value);
-        return TRUE;
-    }
-    return FALSE;
-#else
-    GValue value = { 0 };
-
-    g_return_val_if_fail (result != NULL, FALSE);
-
-    if (ibus_config_get_value (config, section, name, &value)) {
-        *result = g_strdup (g_value_get_string (&value));
-        g_value_unset (&value);
-        return TRUE;
-    }
-    return FALSE;
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
-}
-
-void
-ibus_m17n_config_set_int (IBusConfig  *config,
-                          const gchar *section,
-                          const gchar *name,
-                          gint         value)
-{
-#if IBUS_CHECK_VERSION(1,3,99)
-    ibus_config_set_value (config, section, name, g_variant_new_int32 (value));
-#else
-    GValue v = { 0 };
-
-    g_value_init (&v, G_TYPE_INT);
-    g_value_set_int (&v, value);
-    ibus_config_set_value (config, section, name, &v);
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
-}
-
-gboolean
-ibus_m17n_config_get_int (IBusConfig  *config,
-                          const gchar *section,
-                          const gchar *name,
-                          gint        *result)
-{
-#if IBUS_CHECK_VERSION(1,3,99)
-    GVariant *value = NULL;
-
-    g_return_val_if_fail (result != NULL, FALSE);
-
-    value = ibus_config_get_value (config, section, name);
-    if (value) {
-        *result = g_variant_get_int32 (value);
-        g_variant_unref (value);
-        return TRUE;
-    }
-    return FALSE;
-#else
-    GValue value = { 0 };
-
-    g_return_val_if_fail (result != NULL, FALSE);
-
-    if (ibus_config_get_value (config, section, name, &value)) {
-        *result = g_value_get_int (&value);
-        g_value_unset (&value);
-        return TRUE;
-    }
-    return FALSE;
-#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
-}
-
-#ifdef DEBUG
-#include <locale.h>
-
-int main ()
-{
-    IBusComponent *component;
-    GString *output;
-
-    setlocale (LC_ALL, "");
-    ibus_init ();
-    ibus_m17n_init_common ();
-
-    component = ibus_m17n_get_component ();
-
-    output = g_string_new ("");
-
-    ibus_component_output (component, output, 1);
-
-    g_debug ("\n%s", output->str);
-
-    g_string_free (output, TRUE);
-    g_object_unref (component);
-
-    return 0;
-}
-#endif
